"""
This module is to hold the funcitonality
that converts a thirft service given to us by
thriftpy to our thrift service specification
"""
from collections import defaultdict

from thriftpy.thrift import TType

from thrift_explorer.thrift_models import (
    TI16,
    TI32,
    TI64,
    ServiceEndpoint,
    TBool,
    TByte,
    TDouble,
    TEnum,
    ThriftService,
    ThriftSpec,
    TList,
    TMap,
    TSet,
    TString,
    TStruct,
)

_BASIC_TYPE_MAP = {
    "string": TString,
    "i16": TI16,
    "i32": TI32,
    "i64": TI64,
    "byte": TByte,
    "bool": TBool,
    "double": TDouble,
}


def _parse_type(type_info):
    try:
        ttype_code, nested_type_info = type_info
    except TypeError:
        ttype_code = type_info
        nested_type_info = None

    ttype = TType._VALUES_TO_NAMES[ttype_code].lower()
    if nested_type_info is None:
        return _BASIC_TYPE_MAP[ttype]()
    elif ttype == "list":
        return TList(value_type=_parse_type(nested_type_info))
    elif ttype == "set":
        return TSet(value_type=_parse_type(nested_type_info))
    elif ttype == "map":
        key, value = nested_type_info
        return TMap(key_type=_parse_type(key), value_type=_parse_type(value))
    elif ttype == "struct":
        return TStruct(
            name=nested_type_info.__name__,
            fields=[
                _parse_arg(field_id, result)
                for field_id, result in nested_type_info.thrift_spec.items()
            ],
        )
    # Its a basic type but has defined nested type info. its probably an enum
    return TEnum(
        name=nested_type_info.__name__,
        names_to_values=nested_type_info._NAMES_TO_VALUES,
        values_to_names=nested_type_info._VALUES_TO_NAMES,
    )


def _parse_arg(field_id, thrift_arg):  # Consider renaming?
    try:
        ttype_code, name, required = thrift_arg
        type_info = None
    except ValueError:
        ttype_code, name, type_info, required = thrift_arg
    return ThriftSpec(
        field_id=field_id,
        name=name,
        type_info=_parse_type((ttype_code, type_info)),
        required=required,
    )


def _parse_thrift_endpoint(service, endpoint):
    endpoint_args = getattr(service, "{}_args".format(endpoint))
    endpoint_results = getattr(service, "{}_result".format(endpoint))
    return ServiceEndpoint(
        name=endpoint,
        args=[
            _parse_arg(field_id, arg)
            for field_id, arg in endpoint_args.thrift_spec.items()
        ],
        results=[
            _parse_arg(field_id, result)
            for field_id, result in endpoint_results.thrift_spec.items()
        ],
    )


def _parse_thrift_service(thrift_file, service, endpoints):
    return ThriftService(
        thrift_file,
        service.__name__,
        {endpoint: _parse_thrift_endpoint(service, endpoint) for endpoint in endpoints},
    )


def parse_service_specs(thrifts):
    result = defaultdict(dict)
    for thrift_file, module in thrifts.items():
        for thrift_service in module.__thrift_meta__["services"]:
            result[thrift_file][thrift_service.__name__] = _parse_thrift_service(
                thrift_file,
                thrift_service,
                # I'm a bit confused why this is called services and not 'methods' or 'endpoints'
                thrift_service.thrift_services,
            )
    # Return a standard dict so we can more easily tell when a thrift is not loaded
    return dict(result)



"""
    Here are some things I can theoredically support
    but am punting on until I am further along
    timeout
    certificates/ssl
    timeouts (not sure I want this one)
    http vs rpc (currently supporting just rpc)
    unix socket rather than server
    finagle protocol
    mutliplex protocol
"""
import json
from datetime import timedelta
from enum import Enum, auto

import attr


class CommunicationModelEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, ErrorCode):
            return o.name
        if isinstance(o, Enum):
            return o.value
        if isinstance(o, timedelta):
            return str(o)

        return super().default(o)


class Protocol(Enum):
    BINARY = "tbinaryprotocol"
    JSON = "tjsonprotocol"
    COMPACT = "tcompactprotocol"

    @staticmethod
    def from_string(input_string):
        return Protocol(input_string.lower().strip())


class Transport(Enum):
    BUFFERED = "tbufferedtransport"
    FRAMED = "tframedtransport"

    @staticmethod
    def from_string(input_string):
        return Transport(input_string.lower().strip())


@attr.s(frozen=True)
class ThriftRequest(object):
    """
    The model containing all the information necessary to make a thrift request
        thrift_file: str
            The thrift where the service being called is defined
        service_name: str
            Name of the service being called
        endpoint_name: str
            Endpoint being requested
        host: str
            Hostname (or IP) where the service is hosted
        port: int (or string that is an int)
            Port of the service being hit
        protocol (string, gets converted to a Protocol):
            One of the supported thrift protocols
                BINARY
                JSON
                COMPACT
                MULTIPLEXED
        transport (string, gets converted to a Transport):
            One of the supported thrift transports
                BUFFERED
                FRAMED
        request_body dict:
            dictionary that represents the request being made. Its structure
            is dependent on the request being made
    """

    thrift_file = attr.ib(validator=attr.validators.instance_of(str))
    service_name = attr.ib(validator=attr.validators.instance_of(str))
    endpoint_name = attr.ib(validator=attr.validators.instance_of(str))
    host = attr.ib(validator=attr.validators.instance_of(str))
    port = attr.ib(validator=attr.validators.instance_of(int), converter=int)
    protocol = attr.ib(
        validator=attr.validators.in_(Protocol), converter=Protocol.from_string
    )
    transport = attr.ib(
        validator=attr.validators.in_(Transport), converter=Transport.from_string
    )
    request_body = attr.ib(default=attr.Factory(dict))


@attr.s(frozen=True)
class ThriftResponse(object):
    """
        Model containing the data representing a thrift response

        status: String representing the type of response 'success' or some exception
        request: ThriftRequest used to make this response
        data: dict with the response data
        time_to_make_request: datetime.timedelta Time to make the request
        time_to_connect: datetime.timedelta Time to make the initial connection
    """

    status = attr.ib()
    request = attr.ib()
    data = attr.ib()
    time_to_make_request = attr.ib()
    time_to_connect = attr.ib()


class ErrorCode(Enum):
    THRIFT_NOT_LOADED = auto()
    SERVICE_NOT_IN_THRIFT = auto()
    ENDPOINT_NOT_IN_SERVICE = auto()
    REQUIRED_FIELD_MISSING = auto()
    FIELD_VALIDATION_ERROR = auto()
    INVALID_REQUEST = auto()


@attr.s(frozen=True)
class Error(object):
    code = attr.ib()
    message = attr.ib()


@attr.s(frozen=True)
class FieldError(Error):
    arg_spec = attr.ib()



import sys
from abc import ABC, abstractmethod

import attr


def _validate_basic_type(expected_type, raw_value):
    if isinstance(raw_value, expected_type):
        return None
    return "Expected {0} but got {1}".format(
        expected_type.__name__, type(raw_value).__name__
    )


class ThriftType(ABC):
    def format_arg_for_thrift(self, raw_arg, thrift_module):
        return raw_arg

    @abstractmethod
    def validate_arg(self, raw_arg):
        raise NotImplementedError


@attr.s(frozen=True)
class ThriftService(object):
    """
    Container for a thrift service
    thrift_file: str
        Name of the thrift file this service came from
    name: str
        Name of the service
    endpoints: list[ServiceEndpoint]
        What endpoints can be called on this service
    """

    thrift_file = attr.ib()
    name = attr.ib()
    endpoints = attr.ib()


@attr.s(frozen=True)
class ServiceEndpoint(object):
    """
    Container representing a specific thrift service endpoint
    name: str
        Name of the endpoint
    args: list[ThriftSpec]
        Arguments this service takes in
    results: list[ThriftSpec]
        The result of the method. Its a list because a thrift endpoint
        can have a return value (labeled success) and possibly an exception
        (labeled by the exception name)
    """

    name = attr.ib()
    args = attr.ib()
    results = attr.ib()


@attr.s(frozen=True)
class ThriftSpec(object):
    """
    Container for the specification of an argument
    or a result.
    name: field_id 
        the id of the spec field
    name: str
        Name of the argument/result
    type_info: ThriftType
        If the spec is not a base thrift type this will contain
        info required to specify the complex type
    required: bool
        True if the spec is required in a request/response
        False otherwise
    """

    field_id = attr.ib()
    name = attr.ib()
    type_info = attr.ib()
    required = attr.ib()


@attr.s(frozen=True)
class TStruct(ThriftType):
    """
    Spec for a particular Struct
    name: str
        Name of the struct
    fields: list[ThriftSpec]
        Each property of the struct is its own spec
    ttype: str
        Ttype of the object (always 'struct')
    """

    name = attr.ib()
    fields = attr.ib()
    ttype = attr.ib(default="struct")

    def format_arg_for_thrift(self, raw_arg, clazz):
        class_args = {}
        for field in self.fields:
            thrift_arg = clazz.thrift_spec[field.field_id]
            try:
                ttype_code, name, required = thrift_arg
                type_info = None
            except ValueError:
                ttype_code, name, type_info, required = thrift_arg
            try:
                class_args[field.name] = field.type_info.format_arg_for_thrift(
                    raw_arg[field.name], type_info
                )
            except KeyError:
                continue  # we assume this is already validated. So the arg must be optional
        return clazz(**class_args)

    def validate_arg(self, raw_arg):
        if not isinstance(raw_arg, dict):
            return [
                "Structs expect a dictionary but got {0}".format(type(raw_arg).__name__)
            ]
        errors = []
        for field in self.fields:
            try:
                value = raw_arg[field.name]
                field_validation = field.type_info.validate_arg(value)
                if field_validation:
                    errors.append(
                        "Error with field '{}': '{}'".format(
                            field.name, field_validation
                        )
                    )
            except KeyError:
                if field.required:
                    errors.append("Required field '{}' missing".format(field.name))
        return errors if errors else None


def _validate_collection(collection_class, raw_arg, value_type):
    errors = _validate_basic_type(collection_class, raw_arg)
    if errors:
        return [errors]
    errors = []
    for index, value in enumerate(raw_arg):
        error = value_type.validate_arg(value)
        if error:
            if collection_class == set:
                errors.append("Invalid value in set: {0}".format(error))
            else:
                errors.append("Index {0}: {1}".format(index, error))
    return errors if errors else None


@attr.s(frozen=True)
class TList(ThriftType):
    """
    Spec for a list or a set type
    value_type: ThriftType
        Specification for the type the collection contains
    """

    value_type = attr.ib()
    ttype = attr.ib(default="list")

    def format_arg_for_thrift(self, raw_arg, thrift_module):
        return [
            self.value_type.format_arg_for_thrift(arg, thrift_module) for arg in raw_arg
        ]

    def validate_arg(self, raw_arg):
        return _validate_collection(list, raw_arg, self.value_type)


@attr.s(frozen=True)
class TSet(ThriftType):
    """
    Spec for a list or a set type
    value_type: ThriftType
        Specification for the type the collection contains
    """

    value_type = attr.ib()
    ttype = attr.ib(default="set")

    def format_arg_for_thrift(self, raw_arg, thrift_module):
        return {
            self.value_type.format_arg_for_thrift(arg, thrift_module) for arg in raw_arg
        }

    def validate_arg(self, raw_arg):
        return _validate_collection(set, raw_arg, self.value_type)


@attr.s(frozen=True)
class TMap(ThriftType):
    """
    Spec for a map type
    key_type: ThriftType
        Specification for the type of the key of the map
    value_type: ThriftType
        Specification for the type of the value of the map
    ttype:
        the type (always 'map')
    """

    key_type = attr.ib()
    value_type = attr.ib()
    ttype = attr.ib(default="map")

    def format_arg_for_thrift(self, raw_arg, thrift_module):
        return {
            self.key_type.format_arg_for_thrift(
                key, thrift_module
            ): self.value_type.format_arg_for_thrift(value, thrift_module)
            for key, value in raw_arg.items()
        }

    def validate_arg(self, raw_arg):
        errors = _validate_basic_type(dict, raw_arg)
        if errors:
            return [errors]
        errors = []
        for key, value in raw_arg.items():
            key_validation = self.key_type.validate_arg(key)
            value_validation = self.value_type.validate_arg(value)
            if key_validation:
                errors.append(
                    "Key '{0}' in map invalid: '{1}'".format(key, key_validation)
                )
            if value_validation:
                errors.append(
                    "Value for key '{0}' in map invalid: '{1}'".format(
                        key, value_validation
                    )
                )
        return errors if errors else None


@attr.s(frozen=True)
class TEnum(ThriftType):
    """
    Enums in thrift are a type that holds
    an i32 value that is expected to be one
    of a set of pre_defined values. These
    values each have name associated with them

    name:
        the name of this enum
    names_to_values:
        dict mapping the acceptable names (strings) of this enum
        to their corresponding i32 value
    values_to_names:
        dict mapping the acceptable values (i32s) of this enum
        to their corresponding string value
    ttype:
        the type (always i32)
    """

    name = attr.ib()
    names_to_values = attr.ib()
    values_to_names = attr.ib()
    ttype = attr.ib(default="i32")

    def format_arg_for_thrift(self, raw_arg, _):
        """
        When making the reqeust we want to send the
        int value. If you pass an int we will just return it
        otherwise we will translate the name to its int value.
        """
        try:
            return int(raw_arg)
        except ValueError:
            return self.names_to_values[raw_arg]

    def validate_arg(self, raw_arg):
        if self.names_to_values.get(raw_arg) or self.values_to_names.get(raw_arg):
            return None
        else:
            return "Value is not in enum '{}'".format(self.name)


@attr.s(frozen=True)
class TBool(ThriftType):
    ttype = attr.ib(default="bool")

    def validate_arg(self, raw_arg):
        return _validate_basic_type(bool, raw_arg)


def _numeric_validation(
    raw_arg, python_type, thrift_type_description, min_value, max_value
):
    error = _validate_basic_type(int, raw_arg)
    if error:
        return error
    else:
        if raw_arg > max_value:
            return "Value is too large to be a {}".format(thrift_type_description)
        elif raw_arg < min_value:
            return "Value is too small to be a {}".format(thrift_type_description)
        else:
            return None


@attr.s(frozen=True)
class TByte(ThriftType):
    MIN_VALUE = -128
    MAX_VALUE = 127
    ttype = attr.ib(default="byte")

    def validate_arg(self, raw_arg):
        return _numeric_validation(
            raw_arg, int, "byte", TByte.MIN_VALUE, TByte.MAX_VALUE
        )


@attr.s(frozen=True)
class TI16(ThriftType):
    MIN_VALUE = -32768
    MAX_VALUE = 32767
    ttype = attr.ib(default="i16")

    def validate_arg(self, raw_arg):
        return _numeric_validation(
            raw_arg, int, "16 bit integer", TI16.MIN_VALUE, TI16.MAX_VALUE
        )


@attr.s(frozen=True)
class TI32(ThriftType):
    MIN_VALUE = -2147483648
    MAX_VALUE = 2147483647
    ttype = attr.ib(default="i32")

    def validate_arg(self, raw_arg):
        return _numeric_validation(
            raw_arg, int, "32 bit integer", TI32.MIN_VALUE, TI32.MAX_VALUE
        )


@attr.s(frozen=True)
class TI64(ThriftType):
    MIN_VALUE = -9223372036854775808
    MAX_VALUE = 9223372036854775807
    ttype = attr.ib(default="i64")

    def validate_arg(self, raw_arg):
        return _numeric_validation(
            raw_arg, int, "64 bit integer", TI64.MIN_VALUE, TI64.MAX_VALUE
        )


@attr.s(frozen=True)
class TDouble(ThriftType):
    ttype = attr.ib(default="double")
    # this may bite me some day. sys.float_info implementation dependent.
    # In thrift its 64 bit signed float. That being said if python
    # cant represent the float I dont see what you can dol
    # CPython is always 64 bit, pypy is the same
    MIN_VALUE = sys.float_info.min
    MIN_VALUE = sys.float_info.max

    def validate_arg(self, raw_arg):
        if isinstance(raw_arg, int):
            return None
        return _validate_basic_type(float, raw_arg)


@attr.s(frozen=True)
class TBinary(ThriftType):
    ttype = attr.ib(default="binary")

    def validate_arg(self, raw_arg):
        return _validate_basic_type(bytes, raw_arg)


@attr.s(frozen=True)
class TString(ThriftType):
    ttype = attr.ib(default="string")

    def validate_arg(self, raw_arg):
        return _validate_basic_type(str, raw_arg)



import datetime
import glob
import os
from collections import defaultdict

import thriftpy
from thriftpy import rpc as thriftpy_client
from thriftpy.protocol import (
    TBinaryProtocolFactory,
    TCompactProtocolFactory,
    TJSONProtocolFactory,
)
from thriftpy.thrift import TException

from thrift_explorer.communication_models import (
    Error,
    ErrorCode,
    FieldError,
    Protocol,
    ThriftResponse,
    Transport,
)
from thrift_explorer.thrift_parser import parse_service_specs


def _load_thrifts(thrift_directory):
    thrifts = {}
    thrift_paths = {}
    search_path = os.path.join(thrift_directory, "**/*thrift")
    for thrift_path in glob.iglob(search_path, recursive=True):
        thrift_filename = os.path.basename(thrift_path)
        thrifts[thrift_filename] = thriftpy.load(thrift_path)
        thrift_paths[thrift_filename] = thrift_path
    return thrifts, thrift_paths


def _find_protocol_factory(protocol):
    if protocol == Protocol.BINARY:
        return TBinaryProtocolFactory()
    elif protocol == Protocol.JSON:
        return TJSONProtocolFactory()
    elif protocol == Protocol.COMPACT:
        return TCompactProtocolFactory()
    raise ValueError("Invalid protocol {}".format(protocol))


def _find_transport_factory(transport):
    if transport == Transport.BUFFERED:
        return thriftpy.transport.TBufferedTransportFactory()
    elif transport == Transport.FRAMED:
        return thriftpy.transport.TFramedTransportFactory()
    raise ValueError("Invalid transport {}".format(transport))


def translate_request_body(endpoint, request_body, thriftpy_service_class):
    """
    Translate the request dictionary into whatever arguments
    are required to make the call with the thrift client

    This method assumes the request body had been validated and is correct

    endpoint: The endpoint spec as parsed by ThriftManager

    request_body: dictionary containing the args that should match
    the service. The values should have been validated before hitting
    this method

    thriftpy_service_class: the service class from thriftpy from module created 
     when thriftpy loaded the thrift file
    """
    processed_args = {}
    for arg_spec in endpoint.args:
        thrift_arg = getattr(
            thriftpy_service_class, "{}_args".format(endpoint.name)
        ).thrift_spec[arg_spec.field_id]
        try:
            ttype_code, name, required = thrift_arg
            type_info = None
        except ValueError:
            ttype_code, name, type_info, required = thrift_arg
        try:
            processed_args[arg_spec.name] = arg_spec.type_info.format_arg_for_thrift(
                request_body[arg_spec.name], type_info
            )
        except KeyError:
            # We assume validation happened earlier so if an arg
            # is missing we assume it is not required
            continue
    return processed_args


def translate_thrift_response(response):
    # I considered just using TJSONProtocol. But once it became clear
    # there was not a simple "just call this" method the result
    # started getting as complicated as this.
    if not response:
        return response
    elif isinstance(response, (list,)):
        return [translate_thrift_response(prop) for prop in response]
    elif isinstance(response, (set,)):
        return {translate_thrift_response(prop) for prop in response}
    elif isinstance(response, (dict,)):
        return {
            translate_thrift_response(key): translate_thrift_response(value)
            for key, value in response.items()
        }
    elif hasattr(response, "thrift_spec"):
        struct = {"__thrift_struct_class__": response.__class__.__name__}
        for thrift_spec_parts in response.thrift_spec.values():
            try:
                _, name, _ = thrift_spec_parts
            except ValueError:
                _, name, _, _ = thrift_spec_parts
            struct[name] = translate_thrift_response(getattr(response, name, None))
        return struct
    return response


def find_request_exceptions(thriftpy_service, thrift_request):
    possible_results = getattr(
        thriftpy_service, "{}_result".format(thrift_request.endpoint_name)
    )
    exceptions = []
    for result in possible_results.thrift_spec.values():
        _, _, clazz, _ = result
        if issubclass(clazz, BaseException):
            exceptions.append(clazz)
    return tuple(exceptions)


def _make_client_call(
    client, time_after_client, thrift_request, thriftpy_service, endpoint_spec
):
    translated_request_body = translate_request_body(
        endpoint_spec, thrift_request.request_body, thriftpy_service
    )
    time_before_request = datetime.datetime.now()
    try:
        response = getattr(client, thrift_request.endpoint_name)(
            **translated_request_body
        )
        status = "Success"
    except find_request_exceptions(thriftpy_service, thrift_request) as exception:
        status = exception.__class__.__name__
        response = exception
    except TException as exception:
        status = "ServerError"
        response = "Failed to make call: {}".format(getattr(exception, "message"))
    response_body = translate_thrift_response(response)
    return ThriftResponse(
        status=status,
        request=thrift_request,
        data=response_body,
        time_to_make_request=datetime.datetime.now() - time_before_request,
        time_to_connect=time_after_client,
    )


class ThriftManager(object):
    """
    self.thrift_directory - str - path to the thrift files
    self.thrift_paths - list[str] - list of paths to thrift files
    self.service_spec -  dict[string][dict[string][ThriftService]]

    self.service_spec is keyed by thrift file name. The value is a dictionary 
    keyed by service name with its value being the ThriftService object 
    which has all the useful information you need

    I may need to have a good think about that last field.
    """

    def __init__(self, thrift_directory):
        self.thrift_directory = thrift_directory
        self._thrifts, self.thrift_paths = _load_thrifts(self.thrift_directory)
        self.service_specs = parse_service_specs(self._thrifts)

    def list_thrift_services(self):
        results = defaultdict(list)
        for key in self.service_specs.keys():
            for service in self.service_specs[key]:
                results[key].append(service)
        return results

    def get_thrift(self, thrift):
        return self.service_specs.get(thrift)

    def get_service(self, thrift_name, service_name):
        service = None
        thrift_spec = self.get_thrift(thrift_name)
        if thrift_spec:
            service = self.service_specs[thrift_name].get(service_name)
        return service

    def get_method(self, thrift_name, service_name, method_name):
        method = None
        service = self.get_service(thrift_name, service_name)
        if service:
            method = service.endpoints.get(method_name)
        return method

    def list_methods(self, thrift, service):
        loaded_thrift = self.service_specs[thrift]
        loaded_service = loaded_thrift[service]
        return [key for key in loaded_service.endpoints.keys()]

    def thrift_definition(self, thrift):
        with open(self.thrift_paths[thrift]) as infile:
            return infile.read()

    def _thrift_is_loaded(self, thrift_request):
        try:
            thrift_spec = self.service_specs[thrift_request.thrift_file]
            return None
        except KeyError:
            return [
                Error(
                    code=ErrorCode.THRIFT_NOT_LOADED,
                    message="Thrift File '{}' not loaded in ThriftManager".format(
                        thrift_request.thrift_file
                    ),
                )
            ]

    def _service_in_thrift(self, thrift_spec, service_name, thrift_file):
        try:
            service_spec = thrift_spec[service_name]
        except KeyError:
            return [
                Error(
                    code=ErrorCode.SERVICE_NOT_IN_THRIFT,
                    message="Service '{}' not in thrift '{}'".format(
                        service_name, thrift_file
                    ),
                )
            ]

    def _endpoint_in_service(
        self, service_spec, endpoint_name, service_name, thrift_file
    ):
        try:
            endpoint_spec = service_spec.endpoints[endpoint_name]
        except KeyError:
            return [
                Error(
                    code=ErrorCode.ENDPOINT_NOT_IN_SERVICE,
                    message="Endpoint '{}' not in service '{}' in thrift '{}'".format(
                        endpoint_name, service_name, thrift_file
                    ),
                )
            ]

    def _validate_request_body(
        self, thrift_file, service_name, endpoint_name, request_body
    ):
        validation_errors = []
        for arg_spec in (
            self.service_specs[thrift_file][service_name]
            .endpoints[endpoint_name]
            .args  # this kind of crap makes me think I need to rethink my structures
        ):
            try:
                error = arg_spec.type_info.validate_arg(request_body[arg_spec.name])
                if error:
                    validation_errors.append(
                        FieldError(
                            arg_spec=arg_spec,
                            code=ErrorCode.FIELD_VALIDATION_ERROR,
                            message=error,
                        )
                    )
            except KeyError:
                if arg_spec.required:
                    validation_errors.append(
                        FieldError(
                            arg_spec=arg_spec,
                            code=ErrorCode.REQUIRED_FIELD_MISSING,
                            message="Required Field '{}' not found".format(
                                arg_spec.name
                            ),
                        )
                    )
        return validation_errors

    def validate_request(self, thrift_request):
        return (
            self._thrift_is_loaded(thrift_request)
            or self._service_in_thrift(
                self.service_specs[thrift_request.thrift_file],
                thrift_request.service_name,
                thrift_request.thrift_file,
            )
            or self._endpoint_in_service(
                self.service_specs[thrift_request.thrift_file][
                    thrift_request.service_name
                ],
                thrift_request.endpoint_name,
                thrift_request.service_name,
                thrift_request.thrift_file,
            )
            or self._validate_request_body(
                thrift_request.thrift_file,
                thrift_request.service_name,
                thrift_request.endpoint_name,
                thrift_request.request_body,
            )
        )

    def make_request(self, thrift_request):
        thriftpy_service = getattr(
            self._thrifts[thrift_request.thrift_file], thrift_request.service_name
        )
        thrift_spec = self.service_specs[thrift_request.thrift_file]
        endpoint_spec = thrift_spec[thrift_request.service_name].endpoints[
            thrift_request.endpoint_name
        ]
        time_before_client = datetime.datetime.now()
        try:
            with thriftpy_client.client_context(
                service=thriftpy_service,
                host=thrift_request.host,
                port=thrift_request.port,
                proto_factory=_find_protocol_factory(thrift_request.protocol),
                trans_factory=_find_transport_factory(thrift_request.transport),
            ) as client:
                time_after_client = datetime.datetime.now() - time_before_client
                return _make_client_call(
                    client,
                    time_after_client,
                    thrift_request,
                    thriftpy_service,
                    endpoint_spec,
                )
        except TException as exception:
            status = "ConnectionError"
            response = "Failed to make client connection: {}".format(
                getattr(exception, "message")
            )
            return ThriftResponse(
                status=status,
                request=thrift_request,
                data=response,
                time_to_make_request=None,
                time_to_connect=None,
            )



import json
import os

import attr
from flask import Flask, request

from thrift_explorer.communication_models import (
    CommunicationModelEncoder,
    Error,
    ErrorCode,
    ThriftRequest,
)
from thrift_explorer.thrift_manager import ThriftManager

JSON_CONTENT_TYPE = {"Content-Type": "application/json; charset=utf-8"}
TEXT_CONTENT_TYPE = {"Content-Type": "text/plain; charset=utf-8"}

THRIFT_DIRECTORY_ENV = "THRIFT_DIRECTORY"
DEFAULT_PROTOCOL_ENV = "DEFAULT_THRIFT_PROTOCOL"
DEFAULT_TRANSPORT_ENV = "DEFAULT_THRIFT_TRANSPORT"


def create_app():
    app = Flask(__name__)
    app.config[THRIFT_DIRECTORY_ENV] = os.environ[THRIFT_DIRECTORY_ENV]
    app.config[DEFAULT_PROTOCOL_ENV] = os.environ.get(
        DEFAULT_PROTOCOL_ENV, "TBinaryProtocol"
    )
    app.config[DEFAULT_TRANSPORT_ENV] = os.environ.get(
        DEFAULT_TRANSPORT_ENV, "TBufferedTransport"
    )

    thrift_manager = ThriftManager(app.config[THRIFT_DIRECTORY_ENV])

    def _add_extension_if_needed(thrift):
        if not thrift.endswith(".thrift"):
            thrift = "{}.thrift".format(thrift)
        return thrift

    def _validate_args(thrift, service=None, method=None):
        if not thrift_manager.get_thrift(thrift):
            return "Thrift '{}' not found".format(thrift), 404
        if service and not thrift_manager.get_service(thrift, service):
            return "Service '{}' not found".format(service), 404
        if method and not thrift_manager.get_method(thrift, service, method):
            return "Method '{}' not found".format(method), 404
        return None

    @app.route("/", methods=["GET"])
    def list_services():
        result = []
        for thrift_file, services in thrift_manager.list_thrift_services().items():
            for service in services:
                result.append(
                    {
                        "thrift": thrift_file,
                        "service": service,
                        "methods": sorted(
                            thrift_manager.list_methods(thrift_file, service)
                        ),
                    }
                )
        return (
            json.dumps(
                {"thrifts": sorted(result, key=lambda service: service["thrift"])}
            ),
            200,
            JSON_CONTENT_TYPE,
        )

    @app.route("/<thrift>/", methods=["GET"])
    def get_thrift_definition(thrift):
        thrift = _add_extension_if_needed(thrift)
        error = _validate_args(thrift)
        if error:
            return error
        return thrift_manager.thrift_definition(thrift), 200, TEXT_CONTENT_TYPE

    @app.route("/<thrift>/<service>/", methods=["GET"])
    def get_service_info(thrift, service):
        thrift = _add_extension_if_needed(thrift)
        error = _validate_args(thrift, service)
        if error:
            return error
        methods = thrift_manager.list_methods(thrift, service)
        return (
            json.dumps(
                {"thrift": thrift, "service": service, "methods": sorted(methods)}
            ),
            200,
            JSON_CONTENT_TYPE,
        )

    @app.route("/<thrift>/<service>/<method>/", methods=["GET", "POST"])
    def service_method(thrift, service, method):
        thrift = _add_extension_if_needed(thrift)
        error = _validate_args(thrift, service, method)
        method = thrift_manager.get_method(thrift, service, method)
        if error:
            return error
        if request.method == "POST":
            request_json = request.get_json(force=True)
            errors = []
            try:
                thrift_request = ThriftRequest(
                    thrift_file=thrift,
                    service_name=service,
                    endpoint_name=method.name,
                    host=request_json.get("host"),
                    port=request_json.get("port"),
                    protocol=request_json.get(
                        "protocol", app.config[DEFAULT_PROTOCOL_ENV]
                    ),
                    transport=request_json.get(
                        "transport", app.config[DEFAULT_TRANSPORT_ENV]
                    ),
                    request_body=request_json.get("request_body"),
                )
                errors = thrift_manager.validate_request(thrift_request)
            except ValueError as e:
                errors = [Error(code=ErrorCode.INVALID_REQUEST, message=str(e))]
            except TypeError as e:
                errors = [Error(code=ErrorCode.INVALID_REQUEST, message=str(e.args[0]))]

            if errors:
                return (
                    json.dumps(
                        {
                            "errors": [
                                attr.asdict(error, recurse=True) for error in errors
                            ]
                        },
                        cls=CommunicationModelEncoder,
                    ),
                    400,
                )
            else:
                return (
                    json.dumps(
                        attr.asdict(
                            thrift_manager.make_request(thrift_request), recurse=True
                        ),
                        cls=CommunicationModelEncoder,
                    ),
                    200,
                    JSON_CONTENT_TYPE,
                )
        else:
            return (
                json.dumps(
                    attr.asdict(
                        ThriftRequest(
                            thrift_file=thrift,
                            service_name=service,
                            endpoint_name=method.name,
                            host="<hostname>",
                            port=9090,
                            protocol=app.config[DEFAULT_PROTOCOL_ENV],
                            transport=app.config[DEFAULT_TRANSPORT_ENV],
                            request_body={},
                        ),
                        recurse=True,
                    ),
                    cls=CommunicationModelEncoder,
                ),
                200,
                JSON_CONTENT_TYPE,
            )

    return app
